from __future__ import unicode_literals

import logging

from builtins import str
import six

from lunr.exceptions import BaseLunrException
from lunr.token import Token

log = logging.getLogger(__name__)


class Pipeline:
    """lunr.Pipelines maintain a list of functions to be applied to all tokens
    in documents entering the search index and queries ran agains the index.

    """
    registered_functions = {}

    def __init__(self):
        self._stack = []

    def __len__(self):
        return len(self._stack)

    def __repr__(self):
        return '<Pipeline stack="{}">'.format(
            ','.join(fn.label for fn in self._stack))

    # TODO: add iterator methods?

    @classmethod
    def register_function(cls, fn, label):
        """Register a function with the pipeline."""
        if label in cls.registered_functions:
            log.warning('Overwriting existing registered function %s', label)

        fn.label = label
        cls.registered_functions[fn.label] = fn

    @classmethod
    def load(cls, serialised):
        """Loads a previously serialised pipeline."""
        pipeline = cls()
        for fn_name in serialised:
            try:
                fn = cls.registered_functions[fn_name]
            except KeyError as e:
                raise BaseLunrException(
                    'Cannot load unregistered function '.format(fn_name))
            else:
                pipeline.add(fn)

        return pipeline

    def add(self, *args):
        """Adds new functions to the end of the pipeline.

        Functions must accept three arguments:
        - Token: A lunr.Token object which will be updated
        - i: The index of the token in the set
        - tokens: A list of tokens representing the set
        """
        for fn in args:
            self.warn_if_function_not_registered(fn)
            self._stack.append(fn)

    def warn_if_function_not_registered(self, fn):
        try:
            return fn.label in self.registered_functions
        except AttributeError:
            log.warning(
                'Function "{}" is not registered with pipeline. '
                'This may cause problems when serialising the index.'.format(
                    getattr(fn, 'label', fn)))

    def after(self, existing_fn, new_fn):
        """Adds a single function after a function that already exists in the
        pipeline."""
        self.warn_if_function_not_registered(new_fn)
        try:
            index = self._stack.index(existing_fn)
            self._stack.insert(index + 1, new_fn)
        except ValueError as e:
            six.raise_from(BaseLunrException('Cannot find existing_fn'), e)

    def before(self, existing_fn, new_fn):
        """Adds a single function before a function that already exists in the
        pipeline.

        """
        self.warn_if_function_not_registered(new_fn)
        try:
            index = self._stack.index(existing_fn)
            self._stack.insert(index, new_fn)
        except ValueError as e:
            six.raise_from(BaseLunrException('Cannot find existing_fn'), e)

    def remove(self, fn):
        """Removes a function from the pipeline."""
        try:
            self._stack.remove(fn)
        except ValueError:
            pass

    def run(self, tokens):
        """Runs the current list of functions that make up the pipeline against
        the passed tokens."""
        for fn in self._stack:
            results = []
            for i, token in enumerate(tokens):
                # JS ignores additional arguments to the functions but we
                # force pipeline functions to declare (token, i, tokens)
                # or *args
                result = fn(token, i, tokens)
                if not result:
                    continue
                if isinstance(result, (list, tuple)):  # simulate Array.concat
                    results.extend(result)
                else:
                    results.append(result)
            tokens = results

        return tokens

    def run_string(self, string, metadata=None):
        """Convenience method for passing a string through a pipeline and
        getting strings out. This method takes care of wrapping the passed
        string in a token and mapping the resulting tokens back to strings."""
        token = Token(string, metadata)
        return [str(tkn) for tkn in self.run([token])]

    def reset(self):
        self._stack = []

    def serialize(self):
        return [fn.label for fn in self._stack]



from __future__ import unicode_literals

import re

from lunr.pipeline import Pipeline


def trimmer(token, i=None, tokens=None):
    def trim(s, metadata=None):
        s = re.sub(r'^\W+', '', s)
        s = re.sub(r'\W+$', '', s)
        return s

    return token.update(trim)


Pipeline.register_function(trimmer, 'trimmer')



from __future__ import unicode_literals

from builtins import str

import six

from lunr.tokenizer import Tokenizer


class TestTokenizer:

    def test_splitting_into_tokens(self):
        tokenizer = Tokenizer('foo bar baz')
        tokens = [str(token) for token in tokenizer]

        assert tokens == ['foo', 'bar', 'baz']

    def test_run_downcases_tokens(self):
        tokenizer = Tokenizer('foo BAR BAZ')
        tokens = [str(token) for token in tokenizer]

        assert tokens == ['foo', 'bar', 'baz']

    def test_array_of_strings(self):
        tokenizer = Tokenizer(['foo', 'bar', 'baz'])
        tokens = [str(token) for token in tokenizer]

        assert tokens == ['foo', 'bar', 'baz']

    def test_none_is_converted_to_empty_string(self):
        tokenizer = Tokenizer(['foo', None, 'baz'])
        tokens = [str(token) for token in tokenizer]

        assert tokens == ['foo', '', 'baz']

    def test_multiple_whitespace_is_stripped(self):
        tokenizer = Tokenizer('   foo    bar   baz  ')
        tokens = [str(token) for token in tokenizer]

        assert tokens == ['foo', 'bar', 'baz']

    def test_handling_null_like_arguments(self):
        assert len(Tokenizer(None)) == 0

    def test_converting_a_number_to_tokens(self):
        tokens = [str(token) for token in Tokenizer(41)]
        assert tokens == ['41']

    def test_converting_a_boolean_to_tokens(self):
        tokens = [str(token) for token in Tokenizer(False)]
        assert tokens == ['false']

    def test_converting_an_object_to_tokens(self):
        @six.python_2_unicode_compatible
        class Subject:
            def __str__(self):
                return 'custom object'

        tokens = [str(token) for token in Tokenizer(Subject())]
        assert tokens == ['custom', 'object']

    def test_splits_strings_with_hyphens(self):
        tokens = [str(token) for token in Tokenizer('foo-bar')]
        assert tokens == ['foo', 'bar']

    def test_splits_strings_with_hyphens_and_spaces(self):
        tokens = [str(token) for token in Tokenizer('foo - bar')]
        assert tokens == ['foo', 'bar']

    def test_tracking_the_token_index(self):
        tokens = Tokenizer('foo bar')
        assert tokens[0].metadata['index'] == 0
        assert tokens[1].metadata['index'] == 1

    def test_tracking_the_token_position(self):
        tokens = Tokenizer('foo bar')
        assert tokens[0].metadata['position'] == [0, 3]
        assert tokens[1].metadata['position'] == [4, 3]

    def test_providing_additional_metadata(self):
        tokens = Tokenizer('foo bar', {'hurp': 'durp'})
        assert tokens[0].metadata['hurp'] == 'durp'
        assert tokens[1].metadata['hurp'] == 'durp'



from lunr.utils import CompleteSet


class TestCompleteSet:

    def test_always_contains_other_element(self):
        assert 'foo' in CompleteSet()

    def test_intersection_returns_other(self):
        cs = CompleteSet({'bar'})
        assert cs.intersection({'foo'}) == {'foo'}

    def test_union_returns_other(self):
        cs = CompleteSet({'bar'})
        assert cs.union({'foo'}) == {'foo'}



from __future__ import unicode_literals

import logging

from lunr.__main__ import lunr

__all__ = (lunr,)

logging.basicConfig(format="%(levelname)-7s -  %(message)s")

__VERSION__ = '0.5.1'
__TARGET_JS_VERSION__ = '2.3.2'



from __future__ import unicode_literals
from itertools import chain
from functools import partial

import lunr
from lunr.builder import Builder
from lunr.languages.trimmer import generate_trimmer
from lunr.languages.stemmer import nltk_stemmer, get_language_stemmer
from lunr.pipeline import Pipeline
from lunr.stop_word_filter import stop_word_filter, generate_stop_word_filter

# map from ISO-639-1 codes to SnowballStemmer.languages
# Languages not supported by nltk but by lunr.js: thai, japanese and turkish
# Languages upported by nltk but not lunr.js: arabic

SUPPORTED_LANGUAGES = {
    'ar': 'arabic',
    'da': 'danish',
    'nl': 'dutch',
    'en': 'english',
    'fi': 'finnish',
    'fr': 'french',
    'de': 'german',
    'hu': 'hungarian',
    'it': 'italian',
    'no': 'norwegian',
    'pt': 'portuguese',
    'ro': 'romanian',
    'ru': 'russian',
    'es': 'spanish',
    'sv': 'swedish'
}

try:  # pragma: no cover
    import nltk
    LANGUAGE_SUPPORT = True
except ImportError:  # pragma: no cover
    LANGUAGE_SUPPORT = False


def _get_stopwords_and_word_characters(language):
    nltk.download('stopwords')
    verbose_language = SUPPORTED_LANGUAGES[language]
    stopwords = nltk.corpus.stopwords.words(verbose_language)
    # TODO: search for a more exhaustive list of word characters
    word_characters = {c for word in stopwords for c in word}
    return stopwords, word_characters


def get_nltk_builder(languages):
    """Returns a builder with stemmers for all languages added to it.

    Args:
        languages (list): A list of supported languages.
    """
    all_stemmers = []
    all_stopwords_filters = []
    all_word_characters = set()

    for language in languages:
        if language == 'en':
            # use Lunr's defaults
            all_stemmers.append(lunr.stemmer.stemmer)
            all_stopwords_filters.append(stop_word_filter)
            all_word_characters.update({'\w'})
        else:
            stopwords, word_characters = (
                _get_stopwords_and_word_characters(language))
            all_stemmers.append(
                Pipeline.registered_functions['stemmer-{}'.format(language)])
            all_stopwords_filters.append(
                generate_stop_word_filter(stopwords, language=language))
            all_word_characters.update(word_characters)

    builder = Builder()
    multi_trimmer = generate_trimmer(''.join(sorted(all_word_characters)))
    Pipeline.register_function(
        multi_trimmer, 'lunr-multi-trimmer-{}'.format('-'.join(languages)))
    builder.pipeline.reset()

    for fn in chain([multi_trimmer], all_stopwords_filters, all_stemmers):
        builder.pipeline.add(fn)
    for fn in all_stemmers:
        builder.search_pipeline.add(fn)

    return builder


def register_languages():
    """Register all supported languages to ensure compatibility."""
    for language in set(SUPPORTED_LANGUAGES) - {'en'}:
        language_stemmer = partial(
            nltk_stemmer, get_language_stemmer(language))
        Pipeline.register_function(
            language_stemmer, 'stemmer-{}'.format(language))


if LANGUAGE_SUPPORT:  # pragma: no cover
    # TODO: registering all possible stemmers feels unnecessary but it solves
    # deserializing with arbitrary language functions. Ideally the schema would
    # provide the language(s) for the index and we could register the stemmers
    # as needed
    register_languages()



from __future__ import unicode_literals

from builtins import str

import pytest

from lunr.builder import Builder
from lunr.token_set import TokenSet
from lunr.index import Index
from lunr.vector import Vector


def _assert_deep_keys(dict_, keys):
    d = dict_
    for key in keys.split('.'):
        d_keys_as_str = [str(k) for k in d]
        assert key in d_keys_as_str
        d = d[key]


class TestBuilderBuild:

    def setup_method(self, method):
        self.builder = Builder()
        doc = {
            'id': 'id',
            'title': 'test',
            'body': 'missing'
        }

        self.builder.ref('id')
        self.builder.field('title')
        self.builder.add(doc)
        self.index = self.builder.build()

    def test_adds_tokens_to_inverted_index(self):
        _assert_deep_keys(self.builder.inverted_index, 'test.title.id')

    def test_builds_vector_space_of_the_document_fields(self):
        assert 'title/id' in self.builder.field_vectors
        assert isinstance(self.builder.field_vectors['title/id'], Vector)

    def test_skips_fields_not_defined_for_indexing(self):
        assert 'missing' not in self.builder.inverted_index

    def test_builds_a_token_set_for_the_corpus(self):
        needle = TokenSet.from_string('test')
        assert 'test' in self.builder.token_set.intersect(needle).to_list()

    def test_calculates_document_count(self):
        assert self.builder.average_field_length['title'] == 1

    def test_index_is_returned(self):
        assert isinstance(self.index, Index)


class TestBuilderAdd:

    def test_builder_casts_docrefs_to_strings(self):
        self.builder = Builder()
        self.builder.ref('id')
        self.builder.field('title')

        self.builder.add(dict(id=123, title='test', body='missing'))

        _assert_deep_keys(self.builder.inverted_index, 'test.title.123')

    def test_builder_metadata_whitelist_includes_metadata_in_index(self):
        self.builder = Builder()
        self.builder.ref('id')
        self.builder.field('title')
        self.builder.metadata_whitelist = ['position']

        self.builder.add(dict(id='a', title='test', body='missing'))
        self.builder.add(dict(id='b', title='another test', body='missing'))

        assert self.builder.inverted_index['test']['title']['a'] == {
            'position': [[0, 4]]
        }
        assert self.builder.inverted_index['test']['title']['b'] == {
            'position': [[8, 4]]
        }

    def test_builder_field_raises_if_contains_slash(self):
        self.builder = Builder()

        with pytest.raises(ValueError):
            self.builder.field('foo/bar')

    def test_builder_extracts_nested_properties_from_document(self):
        self.builder = Builder()
        self.builder.field('name', extractor=lambda d: d['person']['name'])

        self.builder.add({
            'id': 'id',
            'person': {
                'name': 'bob'
            }
        })

        assert self.builder.inverted_index['bob']['name']['id'] == {}


class TestBuilderUse:

    def setup_method(self, method):
        self.builder = Builder()

    def test_calls_plugin_function(self):
        def plugin(*args):
            assert True

        self.builder.use(plugin)

    def test_plugin_is_called_with_builder_as_first_argument(self):
        def plugin(builder):
            assert builder is self.builder

        self.builder.use(plugin)

    def test_forwards_arguments_to_the_plugin(self):
        def plugin(builder, *args, **kwargs):
            assert args == (1, 2, 3)
            assert kwargs == {'foo': 'bar'}

        self.builder.use(plugin, 1, 2, 3, foo='bar')


class TestBuilderK1:

    def test_k1_default_value(self):
        builder = Builder()
        assert builder._k1 == 1.2

    def test_k1_can_be_set(self):
        builder = Builder()
        builder.k1(1.6)
        assert builder._k1 == 1.6


class TestBuilderB:

    def test_b_default_value(self):
        builder = Builder()
        assert builder._b == 0.75

    def test_b_within_range(self):
        builder = Builder()
        builder.b(0.5)
        assert builder._b == 0.5

    def test_b_less_than_zero(self):
        builder = Builder()
        builder.b(-1)
        assert builder._b == 0

    def test_b_higher_than_one(self):
        builder = Builder()
        builder.b(1.5)
        assert builder._b == 1


class TestBuilerRef:

    def test_default_reference(self):
        builder = Builder()
        assert builder._ref == 'id'

    def test_defining_a_reference_field(self):
        builder = Builder()
        builder.ref('foo')
        assert builder._ref == 'foo'


class TestBuilderField:

    def test_define_fields_to_index(self):
        builder = Builder()
        builder.field('foo')
        assert len(builder._fields) == 1
        assert builder._fields['foo'].name == 'foo'
        assert builder._fields['foo'].boost == 1
        assert builder._fields['foo'].extractor is None
        assert repr(builder._fields['foo']) == '<Field "foo" boost="1">'
        assert hash(builder._fields['foo']) == hash('foo')



import pytest

from lunr import lunr


@pytest.fixture
def documents():
    return [{
        'id': 'a',
        'title': 'Mr. Green kills Colonel Mustard',
        'body': """Mr. Green killed Colonel Mustard in the study with the
candlestick. Mr. Green is not a very nice fellow.""",
        'word_count': 19
    }, {
        'id': 'b',
        'title': 'Plumb waters plant',
        'body': 'Professor Plumb has a green plant in his study',
        'word_count': 9
    }, {
        'id': 'c',
        'title': 'Scarlett helps Professor',
        'body': """Miss Scarlett watered Professor Plumbs green plant
while he was away from his office last week.""",
        'word_count': 16
    }]


@pytest.fixture
def index(documents):
    return lunr(
        ref='id',
        fields=('title', 'body'),
        documents=documents
    )



from lunr.match_data import MatchData


class TestMatchData:

    def setup_method(self, method):
        self.match = MatchData('foo', 'title', {'position': [1]})
        self.match.combine(
            MatchData('bar', 'title', {'position': [2]}))
        self.match.combine(
            MatchData('baz', 'body', {'position': [3]}))
        self.match.combine(
            MatchData('baz', 'body', {'position': [4]}))

    def test_repr(self):
        assert repr(self.match) == '<MatchData "bar,baz,foo">'

    def test_create_empty_match_data(self):
        assert MatchData().metadata == {}

    def test_create_missing_field(self):
        assert MatchData('foo').metadata['foo'] == {}

    def test_create_missing_metadata(self):
        assert MatchData('foo', 'title').metadata['foo']['title'] == {}

    def test_combine_terms(self):
        assert sorted(
            list(self.match.metadata.keys())) == ['bar', 'baz', 'foo']

    def test_combine_metadata(self):
        assert self.match.metadata['foo']['title']['position'] == [1]
        assert self.match.metadata['bar']['title']['position'] == [2]
        assert self.match.metadata['baz']['body']['position'] == [3, 4]

    def test_combine_does_not_mutate_source_data(self):
        metadata = {'foo': [1]}
        match_data1 = MatchData('foo', 'title', metadata)
        match_data2 = MatchData('foo', 'title', metadata)

        match_data1.combine(match_data2)

        assert metadata['foo'] == [1]

    def test_add_metadata_for_missing_term(self):
        self.match.add('spam', 'title', {'position': [5]})

        assert self.match.metadata['spam']['title']['position'] == [5]

    def test_add_metadata_for_missing_field(self):
        self.match.add('foo', 'body', {'position': [6]})

        assert self.match.metadata['foo']['body']['position'] == [6]

    def test_add_metadata_for_existing_term_field_and_metadata_key(self):
        self.match.add('foo', 'title', {'position': [7]})

        assert self.match.metadata['foo']['title']['position'] == [1, 7]

    def test_add_metadata_for_existing_term_and_field_and_missing_metadata_key(
            self):
        self.match.add('foo', 'title', {'weight': [7]})

        assert self.match.metadata['foo']['title'] == {
            'position': [1],
            'weight': [7],
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest

from lunr import lunr
from lunr.languages import LANGUAGE_SUPPORT, SUPPORTED_LANGUAGES
from lunr.pipeline import Pipeline

documents = [
  {
    "id": "a",
    "text": (
        "Este es un ejemplo inventado de lo que sería un documento en el "
        "idioma que se más se habla en España."),
    "title": "Ejemplo de documento en español"
  },
  {
    "id": "b",
    "text": (
        "Según un estudio que me acabo de inventar porque soy un experto en"
        "idiomas que se hablan en España."),
    "title": "Español es el tercer idioma más hablado del mundo"
  },
]


class TestLanguageSupport:

    @classmethod
    def setup_class(cls):
        assert LANGUAGE_SUPPORT is True, (
            'NLTK not found, please run `pip install -e .[languages]`')

    def test_lunr_function_raises_if_unsupported_language(self):
        with pytest.raises(RuntimeError):
            lunr('id', ['title', 'text'], documents, 'foo')

    def test_lunr_function_raises_if_any_unsupported_language_is_passed(self):
        with pytest.raises(RuntimeError):
            lunr('id', ['title', 'text'], documents, ['es', 'foo'])

    def test_register_languages_in_pipeline_class(self):
        for lang in set(SUPPORTED_LANGUAGES) - {'en'}:
            assert 'stemmer-{}'.format(lang) in Pipeline.registered_functions

    def test_lunr_function_registers_nltk_stemmers_in_pipeline(self):
        idx = lunr('id', ['title', 'text'], documents, ['es', 'it'])
        assert 'stemmer-es' in repr(idx.pipeline)
        assert 'stemmer-it' in repr(idx.pipeline)

    def test_lunr_registers_lun_stemmers_in_pipeline_if_language_is_en(self):
        idx = lunr('id', ['title', 'text'], documents, ['en', 'es'])
        assert 'stemmer,stemmer-es' in repr(idx.pipeline)

    def test_search_stems_search_terms(self):
        idx = lunr('id', ['title', 'text'], documents, 'es')
        results = idx.search('inventando')  # stemmed to "invent"
        assert len(results) == 2

    def test_search_stems_search_terms_for_both_languages(self):
        italian_document = {
            "id": "c",
            "text": (
                "Secondo uno studio che ho appena inventato perché sono un "
                "esperto di lingue parlate in Spagna."),
            "title": "Lo spagnolo è la terza lingua più parlata al mondo"
        }
        idx = lunr(
            ref='id',
            fields=['title', 'text'],
            documents=(documents + [italian_document]),
            languages=['es', 'it'])
        results = idx.search('spagna')
        assert len(results) == 1

        results = idx.search('inventando')
        assert len(results) == 2



from __future__ import unicode_literals

from builtins import str


def as_string(obj):
    return '' if not obj else str(obj)


class CompleteSet(set):

    def union(self, other):
        return set(other)

    def intersection(self, other):
        return set(other)

    def __contains__(self, y):
        return True



from __future__ import unicode_literals

from builtins import str

from copy import deepcopy
import re

from lunr.token import Token
from lunr.utils import as_string

SEPARATOR = r'[\s\-]+'


def Tokenizer(obj, metadata=None):
    """Splits a string into tokens ready to be inserted into the search index.

    Uses `Tokenizer.SEPARATOR` to split strings, change the value of this
    property to change how strings are split into tokens.

    This tokenizer will convert its parameter to a string by calling `str` and
    then will split this string on the character in `Tokenizer.SEPARATOR`.
    Lists will have their elements converted to strings and wrapped in a lunr
    `Token`.

    Optional metadata can be passed to the tokenizer, this metadata will be
    cloned and added as metadata to every token that is created from the object
    to be tokenized.
    """
    if obj is None:
        return []

    metadata = metadata or {}

    if isinstance(obj, (list, tuple)):
        return [
            Token(as_string(element).lower(), deepcopy(metadata))
            for element in obj]

    string = str(obj).strip().lower()
    length = len(string)
    tokens = []
    slice_start = 0
    for slice_end in range(length):
        char = string[slice_end]
        slice_length = slice_end - slice_start
        if re.match(SEPARATOR, char) or slice_end == length - 1:
            if slice_length > 0:
                sl = slice(
                    slice_start,
                    slice_end if slice_end < length - 1 else None)

                token_metadata = deepcopy(metadata)
                token_metadata['position'] = [
                    slice_start,
                    slice_length
                    if slice_end < length - 1 else slice_length + 1]
                token_metadata['index'] = len(tokens)

                tokens.append(Token(string[sl], token_metadata))

            slice_start = slice_end + 1

    return tokens



import pytest

from lunr.token_set import TokenSet
from lunr.exceptions import BaseLunrException


class TestTokenSetStr:

    def test_str_includes_node_finality(self):
        non_final = TokenSet()
        final = TokenSet()
        other_final = TokenSet()

        final.final = True
        other_final.final = True

        assert str(non_final) != str(final)
        assert str(other_final) == str(final)

    def test_str_includes_all_edges(self):
        zero_edges = TokenSet()
        one_edge = TokenSet()
        two_edges = TokenSet()

        one_edge.edges['a'] = 1
        two_edges.edges['a'] = 1
        two_edges.edges['b'] = 1

        assert str(zero_edges) != str(one_edge)
        assert str(two_edges) != str(one_edge)
        assert str(two_edges) != str(zero_edges)

    def test_str_includes_edge_id(self):
        child_a = TokenSet()
        child_b = TokenSet()
        parent_a = TokenSet()
        parent_b = TokenSet()
        parent_c = TokenSet()

        parent_a.edges['a'] = child_a
        parent_b.edges['a'] = child_b
        parent_c.edges['a'] = child_b

        assert str(parent_b) == str(parent_c)
        assert str(parent_a) != str(parent_c)
        assert str(parent_a) != str(parent_b)


class TestTokenSetFromString:

    def test_from_string_without_wildcard(self):
        TokenSet._next_id = 1
        x = TokenSet.from_string('a')

        assert str(x) == '0a2'
        assert x.edges['a'].final

    def test_from_string_with_trailing_wildcard(self):
        x = TokenSet.from_string('a*')
        wild = x.edges['a'].edges['*']

        assert wild == wild.edges['*']
        assert wild.final


class TestTokenSetFromList:

    def test_from_list_with_unsorted_list(self):
        with pytest.raises(BaseLunrException):
            TokenSet.from_list(['z', 'a'])

    def test_from_list_with_sorted_list(self):
        token_set = TokenSet.from_list(['a', 'z'])
        assert ['a', 'z'] == sorted(token_set.to_list())

    def test_from_list_is_minimal(self):
        token_set = TokenSet.from_list(['ac', 'dc'])
        ac_node = token_set.edges['a'].edges['c']
        dc_node = token_set.edges['d'].edges['c']

        assert ac_node == dc_node


class TestTokenSetToList:

    def test_to_list_includes_all_words(self):
        words = ['bat', 'cat']
        token_set = TokenSet.from_list(words)

        assert set(words) == set(token_set.to_list())

    def test_to_list_includes_single_words(self):
        word = 'bat'
        token_set = TokenSet.from_string(word)

        assert set([word]) == set(token_set.to_list())


class TestTokenSetIntersect:

    def test_intersect_no_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('bar')
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_intersect_simple_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('cat')
        z = x.intersect(y)

        assert {'cat'} == set(z.to_list())

    def test_intersect_trailing_wildcard_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('c*')
        z = x.intersect(y)

        assert {'cat'} == set(z.to_list())

    def test_intersect_trailing_wildcard_no_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('b*')
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_intersect_leading_wildcard_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('*t')
        z = x.intersect(y)

        assert {'cat'} == set(z.to_list())

    def test_intersect_leading_wildcard_no_intersection(self):
        x = TokenSet.from_string('cat')
        y = TokenSet.from_string('*r')
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_intersect_contained_wildcard_intersection(self):
        x = TokenSet.from_string('foo')
        y = TokenSet.from_string('f*o')
        z = x.intersect(y)

        assert {'foo'} == set(z.to_list())

    def test_intersect_contained_wildcard_no_intersection(self):
        x = TokenSet.from_string('foo')
        y = TokenSet.from_string('b*r')
        z = x.intersect(y)

        assert len(z.to_list()) == 0

    def test_intersect_wildcard_zero_or_more_characters(self):
        x = TokenSet.from_string('foo')
        y = TokenSet.from_string('foo*')
        z = x.intersect(y)

        assert {'foo'} == set(z.to_list())

    def test_intersect_with_fuzzy_string_subtitution(self):
        x1 = TokenSet.from_string('bar')
        x2 = TokenSet.from_string('cur')
        x3 = TokenSet.from_string('cat')
        x4 = TokenSet.from_string('car')
        x5 = TokenSet.from_string('foo')
        y = TokenSet.from_fuzzy_string('car', 1)

        assert x1.intersect(y).to_list() == ['bar']
        assert x2.intersect(y).to_list() == ['cur']
        assert x3.intersect(y).to_list() == ['cat']
        assert x4.intersect(y).to_list() == ['car']
        assert x5.intersect(y).to_list() == []

    def test_intersect_with_fuzzy_string_deletion(self):
        x1 = TokenSet.from_string('ar')
        x2 = TokenSet.from_string('br')
        x3 = TokenSet.from_string('ba')
        x4 = TokenSet.from_string('bar')
        x5 = TokenSet.from_string('foo')
        y = TokenSet.from_fuzzy_string('bar', 1)

        assert x1.intersect(y).to_list() == ['ar']
        assert x2.intersect(y).to_list() == ['br']
        assert x3.intersect(y).to_list() == ['ba']
        assert x4.intersect(y).to_list() == ['bar']
        assert x5.intersect(y).to_list() == []

    def test_intersect_with_fuzzy_string_insertion(self):
        x1 = TokenSet.from_string('bbar')
        x2 = TokenSet.from_string('baar')
        x3 = TokenSet.from_string('barr')
        x4 = TokenSet.from_string('bar')
        x5 = TokenSet.from_string('ba')
        x6 = TokenSet.from_string('foo')
        y = TokenSet.from_fuzzy_string('bar', 1)

        assert x1.intersect(y).to_list() == ['bbar']
        assert x2.intersect(y).to_list() == ['baar']
        assert x3.intersect(y).to_list() == ['barr']
        assert x4.intersect(y).to_list() == ['bar']
        assert x5.intersect(y).to_list() == ['ba']
        assert x6.intersect(y).to_list() == []

    def test_intersect_with_fuzzy_string_transpose(self):
        x1 = TokenSet.from_string('abr')
        x2 = TokenSet.from_string('bra')
        x3 = TokenSet.from_string('foo')
        y = TokenSet.from_fuzzy_string('bar', 1)

        assert x1.intersect(y).to_list() == ['abr']
        assert x2.intersect(y).to_list() == ['bra']
        assert x3.intersect(y).to_list() == []



from __future__ import unicode_literals

import re


def generate_trimmer(word_characters):
    """Returns a trimmer function from a string of word characters.

    TODO: lunr-languages ships with lists of word characters for each language
    I haven't found an equivalent in Python, we may need to copy it.
    """
    start_re = r'^[^{}]+'.format(word_characters)
    end_re = r'[^{}]+$'.format(word_characters)

    def trimmer(token, i=None, tokens=None):
        def trim(s, metadata=None):
            s = re.sub(start_re, '', s)
            s = re.sub(end_re, '', s)
            return s

        return token.update(trim)

    return trimmer



import json

from lunr import lunr
from lunr.index import Index


class TestSerialization:

    def setup_method(self, method):
        documents = [{
            'id': 'a',
            'title': 'Mr. Green kills Colonel Mustard',
            'body': """Mr. Green killed Colonel Mustard in the study with the
candlestick. Mr. Green is not a very nice fellow.""",
            'word_count': 19
        }, {
            'id': 'b',
            'title': 'Plumb waters plant',
            'body': 'Professor Plumb has a green plant in his study',
            'word_count': 9
        }, {
            'id': 'c',
            'title': 'Scarlett helps Professor',
            'body': """Miss Scarlett watered Professor Plumbs green plant
while he was away from his office last week.""",
            'word_count': 16
        }]

        self.idx = lunr(
            ref='id',
            fields=('title', 'body'),
            documents=documents
        )

    def test_serialization(self):
        serialized_index = json.dumps(self.idx.serialize())
        loaded_index = Index.load(json.loads(serialized_index))

        assert self.idx.search('green') == loaded_index.search('green')



import pytest

from lunr.query import Query, QueryPresence
from lunr.token import Token
from lunr.tokenizer import Tokenizer


class BaseQuerySuite:
    ALL_FIELDS = ['title', 'body']

    def setup_method(self, method):
        self.query = Query(self.ALL_FIELDS)


class TestQueryTerm(BaseQuerySuite):

    def test_single_string_term_adds_single_clause(self):
        self.query.clause(term='foo')

        assert len(self.query.clauses) == 1
        assert self.query.clauses[0].term == 'foo'
        assert repr(self.query) == '<Query fields="title,body" clauses="foo">'
        assert repr(self.query.clauses[0]) == '<Clause term="foo">'

    def test_single_token_term_adds_single_clause(self):
        self.query.term(Token('foo'))

        assert len(self.query.clauses) == 1
        assert self.query.clauses[0].term == 'foo'

    def test_multiple_string_terms_adds_multiple_clauses(self):
        self.query.term(['foo', 'bar'])

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == 'foo'
        assert self.query.clauses[1].term == 'bar'
        assert repr(self.query) == (
            '<Query fields="title,body" clauses="foo,bar">')

    def test_multiple_token_terms_adds_multiple_clauses(self):
        self.query.term(Tokenizer('foo bar'))

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == 'foo'
        assert self.query.clauses[1].term == 'bar'

    def test_multiple_string_terms_with_options(self):
        self.query.term(['foo', 'bar'], use_pipeline=False)

        assert len(self.query.clauses) == 2
        assert self.query.clauses[0].term == 'foo'
        assert self.query.clauses[1].term == 'bar'


class TestQueryClause(BaseQuerySuite):

    def test_clause_defaults(self):
        self.query.clause(term='foo')
        self.clause = self.query.clauses[0]

        assert self.clause.fields == self.ALL_FIELDS
        assert self.clause.boost == 1
        assert self.clause.use_pipeline is True

    def test_clause_specified(self):
        self.query.clause(
            term='foo',
            boost=10,
            fields=['title'],
            use_pipeline=False
        )
        self.clause = self.query.clauses[0]

        assert self.clause.fields == ['title']
        assert self.clause.boost == 10
        assert self.clause.use_pipeline is False

    @pytest.mark.parametrize(
        'wildcard, expected_term', [
            (Query.WILDCARD_NONE, 'foo'),
            (Query.WILDCARD_LEADING, '*foo'),
            (Query.WILDCARD_TRAILING, 'foo*'),
            (Query.WILDCARD_LEADING | Query.WILDCARD_TRAILING, '*foo*'),
        ])
    def test_clause_wildcard(self, wildcard, expected_term):
        self.query.clause(term='foo', wildcard=wildcard)
        self.clause = self.query.clauses[0]

        assert self.clause.term == expected_term

    def test_clause_wildcard_existing(self):
        self.query.clause(
            term='*foo*',
            wildcard=Query.WILDCARD_LEADING | Query.WILDCARD_TRAILING
        )
        self.clause = self.query.clauses[0]

        assert self.clause.term == '*foo*'


class TestQueryIsNegated(BaseQuerySuite):

    def test_all_prohibited(self):
        self.query.term('foo', presence=QueryPresence.PROHIBITED)
        self.query.term('bar', presence=QueryPresence.PROHIBITED)

        assert self.query.is_negated() is True

    def test_some_prohibited(self):
        self.query.term('foo', presence=QueryPresence.PROHIBITED)
        self.query.term('bar', presence=QueryPresence.REQUIRED)

        assert self.query.is_negated() is False

    def test_nome_prohibited(self):
        self.query.term('foo', presence=QueryPresence.OPTIONAL)
        self.query.term('bar', presence=QueryPresence.REQUIRED)

        assert self.query.is_negated() is False



from __future__ import unicode_literals

from builtins import str

import six


@six.python_2_unicode_compatible
class Token:

    def __init__(self, string='', metadata=None):
        self.string = string
        self.metadata = metadata or {}

    def __str__(self):
        return self.string

    def __repr__(self):
        return '<Token "{}">'.format(str(self))

    def update(self, fn):
        """A token update function is used when updating or optionally
        when cloning a token."""
        # TODO: we require functions to have two parameters, JS doesn't care
        self.string = fn(self.string, self.metadata)
        return self

    def clone(self, fn=None):
        """Applies the given function to the wrapped string token."""
        fn = fn or (lambda s, m: s)
        return Token(fn(self.string, self.metadata), self.metadata)



from mock import patch

import pytest

from lunr.exceptions import BaseLunrException
from lunr.pipeline import Pipeline


def noop(*args, **kwargs):
    pass


def fn(*args, **kwargs):
    pass


class BaseTestPipeline:

    @pytest.fixture(autouse=True)
    def setup_mock_pipline(self, monkeypatch):
        monkeypatch.setattr(Pipeline, 'registered_functions', {})
        monkeypatch.setattr(Pipeline, 'warn_if_function_not_registered', noop)
        self.pipeline = Pipeline()


class TestAdd(BaseTestPipeline):

    def test_add_function_to_pipeline(self):
        self.pipeline.add(noop)
        assert len(self.pipeline) == 1

    def test_add_multiple_functions_to_pipeline(self):
        self.pipeline.add(noop, noop)
        assert len(self.pipeline) == 2

    def test_add_warns_if_function_not_registered(self, monkeypatch):
        monkeypatch.undo()
        with patch('lunr.pipeline.log') as mock_log:
            self.pipeline.add(lambda x: x)
            mock_log.warning.assert_called_once()


class TestRemove(BaseTestPipeline):

    def test_remove_function_exists_in_pipeline(self):
        self.pipeline.add(noop)
        assert len(self.pipeline) == 1

        self.pipeline.remove(noop)
        assert len(self.pipeline) == 0

    def test_remove_function_does_not_exist_in_pipeline(self):

        self.pipeline.add(noop)
        assert len(self.pipeline) == 1

        self.pipeline.remove(fn)
        assert len(self.pipeline) == 1


class TestBefore(BaseTestPipeline):

    def test_before_other_function_exists(self):
        self.pipeline.add(noop)
        self.pipeline.before(noop, fn)

        assert self.pipeline._stack == [fn, noop]

    def test_before_other_functions_does_not_exist(self):
        with pytest.raises(BaseLunrException):
            self.pipeline.before(noop, fn)

        assert len(self.pipeline) == 0


class TestAfter(BaseTestPipeline):

    def test_after_other_function_exists(self):
        self.pipeline.add(noop)
        self.pipeline.after(noop, fn)

        assert self.pipeline._stack == [noop, fn]

    def test_after_other_function_does_not_exist(self):
        with pytest.raises(BaseLunrException):
            self.pipeline.after(noop, fn)

        assert len(self.pipeline) == 0


class TestRun(BaseTestPipeline):

    def test_run_calling_each_function_for_each_token(self):
        count_1 = []
        count_2 = []

        def fn1(t, *args):
            count_1.append(1)
            return t

        def fn2(t, *args):
            count_2.append(1)
            return t

        self.pipeline.add(fn1, fn2)
        self.pipeline.run([1, 2, 3])

        assert len(count_1) == 3
        assert len(count_2) == 3

    def test_run_passes_token_to_pipeline_function(self):
        def fn(token, *args):
            assert token == 'foo'

        self.pipeline.add(fn)
        self.pipeline.run(['foo'])

    def test_run_passes_index_to_pipeline_function(self):
        def fn(_, index, *args):
            assert index == 0

        self.pipeline.add(fn)
        self.pipeline.run(['foo'])

    def test_run_passes_entire_token_list_to_pipeline_function(self):
        def fn(_, __, tokens):
            assert tokens == ['foo']

        self.pipeline.add(fn)
        self.pipeline.run(['foo'])

    def test_run_passes_output_of_one_function_as_input_to_the_next(self):
        def fn1(t, *args):
            return t.upper()

        def fn2(t, *args):
            assert t == 'FOO'

        self.pipeline.add(fn1, fn2)
        self.pipeline.run(['foo'])

    def test_run_returns_the_results_of_the_last_function(self):
        def fn(t, *args):
            return t.upper()

        self.pipeline.add(fn)

        assert self.pipeline.run(['foo']) == ['FOO']

    def test_run_filters_out_none_values(self):
        tokens = []

        def fn1(t, i, _):
            if i % 2:
                return t

        def fn2(t, *args):
            tokens.append(t)
            return t

        self.pipeline.add(fn1)
        self.pipeline.add(fn2)

        output = self.pipeline.run(list('abcd'))

        assert tokens == ['b', 'd']
        assert output == ['b', 'd']

    def test_expanding_tokens_passed_to_output(self):
        self.pipeline.add(lambda t, *args: [t, t.upper()])

        assert self.pipeline.run(['foo']) == ['foo', 'FOO']

    def test_expanding_tokens_not_passed_to_same_function(self):
        received = []

        def fn(t, *args):
            received.append(t)
            return [t, t.upper()]

        self.pipeline.add(fn)
        self.pipeline.run(['foo'])

        assert received == ['foo']

    def test_expanding_tokens_passed_to_the_next_pipeline_function(self):
        received = []

        def fn1(t, *args):
            return [t, t.upper()]

        def fn2(t, *args):
            received.append(t)

        self.pipeline.add(fn1)
        self.pipeline.add(fn2)
        self.pipeline.run(['foo'])

        assert received == ['foo', 'FOO']


class TestSerialize(BaseTestPipeline):

    def test_serialize_returns_array_of_registered_function_labels(self):
        Pipeline.register_function(fn, 'fn')
        self.pipeline.add(fn)

        assert self.pipeline.serialize() == ['fn']
        assert repr(self.pipeline) == '<Pipeline stack="fn">'


class TestRegisterFunction(BaseTestPipeline):

    def setup_method(self, method):
        def fn(*args):
            pass

        self.fn = fn

    def test_register_function_adds_a_label_property_to_the_function(self):
        Pipeline.register_function(self.fn, 'fn')

        assert self.fn.label == 'fn'

    def test_register_function_adds_function_to_list_of_registered_functions(
            self):
        Pipeline.register_function(self.fn, 'fn')

        assert Pipeline.registered_functions['fn'] == self.fn

    def test_register_function_warns_when_adding_function_with_same_label(
            self):
        Pipeline.register_function(self.fn, 'fn')
        with patch('lunr.pipeline.log') as mock_log:
            Pipeline.register_function(self.fn, 'fn')

            mock_log.warning.assert_called_once()


class TestLoad(BaseTestPipeline):

    def test_load_with_registered_functions(self):
        serialized_pipeline = ['fn']
        Pipeline.register_function(fn, 'fn')

        pipeline = Pipeline.load(serialized_pipeline)

        assert len(pipeline) == 1
        assert pipeline._stack[0] == fn

    def test_load_with_unregistered_functions(self):
        serialized_pipeline = ['fn']
        with pytest.raises(BaseLunrException):
            Pipeline.load(serialized_pipeline)


class TestReset(BaseTestPipeline):

    def test_reset_empties_the_stack(self):
        self.pipeline.add(noop)
        assert len(self.pipeline) == 1

        self.pipeline.reset()
        assert len(self.pipeline) == 0



from lunr.stop_word_filter import stop_word_filter, generate_stop_word_filter
from lunr.pipeline import Pipeline

STOP_WORDS = ['the', 'and', 'but', 'than', 'when']


class TestStopWordFilter:

    def test_filters_stop_words(self):
        for word in STOP_WORDS:
            assert stop_word_filter(word) is None

    def test_ignores_non_stop_words(self):
        non_stop_words = ['interesting', 'words', 'pass', 'through']
        for word in non_stop_words:
            assert stop_word_filter(word) == word

    def test_is_a_registered_pipeline_function(self):
        assert stop_word_filter.label == 'stopWordFilter'
        assert Pipeline.registered_functions[
            'stopWordFilter'] == stop_word_filter


class TestGenerateStopWordFilter:

    def test_creates_correct_stop_words_filter(self):
        new_stop_word_filter = generate_stop_word_filter(STOP_WORDS)
        for word in STOP_WORDS:
            assert new_stop_word_filter(word) is None

    def test_registers_new_stop_words_filter(self):
        new_stop_word_filter = generate_stop_word_filter(STOP_WORDS)
        assert new_stop_word_filter.label == 'stopWordFilter'
        assert Pipeline.registered_functions[
            'stopWordFilter'] == new_stop_word_filter

    def test_passing_a_language_adds_to_registered_label(self):
        new_stop_word_filter = generate_stop_word_filter(STOP_WORDS, 'es')
        assert new_stop_word_filter.label == 'stopWordFilter-es'
        assert Pipeline.registered_functions[
            'stopWordFilter-es'] == new_stop_word_filter



import os.path

import pytest

from tests.utils import read_json_fixture

from lunr import lunr
from lunr.pipeline import Pipeline


def index_mkdocs_data(data):
    lunr(
        ref='id',
        fields=('title', 'text'),
        documents=data['docs']
    )


def test_index_mkdocs(benchmark):
    data = read_json_fixture('mkdocs_index.json')
    benchmark(index_mkdocs_data, data)


class TestPipelineBenchmarks:

    FEW_COUNT = 50
    MANY_COUNT = 1000

    @pytest.fixture(scope='session')
    def many_tokens(self):
        path = os.path.join(os.path.dirname(__file__), 'fixtures/words.txt')
        with open(path) as words:
            self.many_tokens = [
                words.readline().strip() for _ in range(self.MANY_COUNT)]
        self.few_tokens = self.many_tokens[:self.FEW_COUNT]
        yield self.many_tokens

    @pytest.fixture(scope='session')
    def few_tokens(self, many_tokens):
        yield self.few_tokens

    @staticmethod
    def token_to_token(token, i, tokens):
        return token

    @staticmethod
    def token_to_token_array(token, i, tokens):
        return [token, token]

    def test_few_token_to_token(self, few_tokens, benchmark):
        token_to_token_pipeline = Pipeline()
        token_to_token_pipeline.add(self.token_to_token)
        benchmark(token_to_token_pipeline.run, few_tokens)

    def test_many_token_to_token(self, many_tokens, benchmark):
        token_to_token_pipeline = Pipeline()
        token_to_token_pipeline.add(self.token_to_token)
        benchmark(token_to_token_pipeline.run, many_tokens)

    def test_few_token_to_token_array(self, few_tokens, benchmark):
        token_to_token_array_pipeline = Pipeline()
        token_to_token_array_pipeline.add(self.token_to_token_array)
        benchmark(token_to_token_array_pipeline.run, few_tokens)

    def test_many_token_to_token_array(self, many_tokens, benchmark):
        token_to_token_array_pipeline = Pipeline()
        token_to_token_array_pipeline.add(self.token_to_token_array)
        benchmark(token_to_token_array_pipeline.run, many_tokens)



from __future__ import unicode_literals

from builtins import str

import pytest

from lunr.trimmer import trimmer
from lunr.token import Token
from lunr.pipeline import Pipeline


class TestTrimmer:

    def test_latin_characters(self):
        token = Token('hello')
        assert str(trimmer(token)) == str(token)

    @pytest.mark.parametrize('description, string, expected', [
            ('full stop', 'hello.', 'hello'),
            ('inner apostrophe', "it's", "it's"),
            ('trailing apostrophe', "james'", 'james'),
            ('exclamation mark', 'stop!', 'stop'),
            ('comma', 'first,', 'first'),
            ('brackets', '[tag]', 'tag'),
        ])
    def test_punctuation(self, description, string, expected):
        token = Token(string)
        trimmed = str(trimmer(token))

        assert trimmed == expected

    def test_is_a_registered_pipeline_function(self):
        assert trimmer.label == 'trimmer'
        assert Pipeline.registered_functions['trimmer'] == trimmer



from __future__ import unicode_literals, division

from math import sqrt

from lunr.exceptions import BaseLunrException


class Vector:
    """A vector is used to construct the vector space of documents and queries.
    These vectors support operations to determine the similarity between two
    documents or a document and a query.

    Normally no parameters are required for initializing a vector, but in the
    case of loading a previously dumped vector the raw elements can be provided
    to the constructor.

    For performance reasons vectors are implemented with a flat array, where an
    elements index is immediately followed by its value.
    E.g. [index, value, index, value].

    TODO: consider implemetation as 2-tuples.

    This allows the underlying array to be as sparse as possible and still
    offer decent performance when being used for vector calculations.
    """

    def __init__(self, elements=None):
        self._magnitude = 0
        self.elements = elements or []

    def __repr__(self):
        return '<Vector magnitude={}>'.format(self.magnitude)

    def __iter__(self):
        return iter(self.elements)

    def position_for_index(self, index):
        """Calculates the position within the vector to insert a given index.

        This is used internally by insert and upsert. If there are duplicate
        indexes then the position is returned as if the value for that index
        were to be updated, but it is the callers responsibility to check
        whether there is a duplicate at that index
        """
        if not self.elements:
            return 0

        start = 0
        end = int(len(self.elements) / 2)
        slice_length = end - start
        pivot_point = int(slice_length / 2)
        pivot_index = self.elements[pivot_point * 2]

        while slice_length > 1:
            if pivot_index < index:
                start = pivot_point
            elif pivot_index > index:
                end = pivot_point
            else:
                break

            slice_length = end - start
            pivot_point = start + int(slice_length / 2)
            pivot_index = self.elements[pivot_point * 2]

        if pivot_index == index:
            return pivot_point * 2
        elif pivot_index > index:
            return pivot_point * 2
        else:
            return (pivot_point + 1) * 2

    def insert(self, insert_index, val):
        """Inserts an element at an index within the vector.

        Does not allow duplicates, will throw an error if there is already an
        entry for this index.
        """
        def prevent_duplicates(index, val):
            raise BaseLunrException('Duplicate index')
        self.upsert(insert_index, val, prevent_duplicates)

    def upsert(self, insert_index, val, fn=None):
        """Inserts or updates an existing index within the vector.

        Args:
            - insert_index (int): The index at which teh element should be
                inserted.
            - val (int|float): The value to be inserted into the vector.
            - fn (callable, optional): An optional callable taking two
                arguments, the current value and the passed value to generate
                the final inserted value at the position in case of collision.
        """
        fn = fn or (lambda current, passed: passed)
        self._magnitude = 0
        position = self.position_for_index(insert_index)
        if (position < len(self.elements) and
                self.elements[position] == insert_index):
            self.elements[position + 1] = fn(self.elements[position + 1], val)
        else:
            self.elements.insert(position, val)
            self.elements.insert(position, insert_index)

    def to_list(self):
        """Converts the vector to an array of the elements within the vector"""
        output = []
        for i in range(1, len(self.elements), 2):
            output.append(self.elements[i])
        return output

    def serialize(self):
        # TODO: the JS version forces rounding on the elements upon insertion
        # to ensure symmetry upon serialization
        return [round(element, 3) for element in self.elements]

    @property
    def magnitude(self):
        if not self._magnitude:
            sum_of_squares = 0
            for i in range(1, len(self.elements), 2):
                value = self.elements[i]
                sum_of_squares += value * value

            self._magnitude = sqrt(sum_of_squares)

        return self._magnitude

    def dot(self, other):
        """Calculates the dot product of this vector and another vector."""
        dot_product = 0
        a = self.elements
        b = other.elements
        a_len = len(a)
        b_len = len(b)
        i = j = 0

        while i < a_len and j < b_len:
            a_val = a[i]
            b_val = b[j]
            if a_val < b_val:
                i += 2
            elif a_val > b_val:
                j += 2
            else:
                dot_product += a[i + 1] * b[j + 1]
                i += 2
                j += 2

        return dot_product

    def similarity(self, other):
        """Calculates the cosine similarity between this vector and another
        vector."""
        if self.magnitude == 0 or other.magnitude == 0:
            return 0

        return self.dot(other) / self.magnitude



from __future__ import unicode_literals

import six

from lunr.exceptions import BaseLunrException


@six.python_2_unicode_compatible
class FieldRef:

    JOINER = '/'

    def __init__(self, doc_ref, field_name, string_value=None):
        self.doc_ref = doc_ref
        self.field_name = field_name
        self._string_value = string_value

    def __repr__(self):
        return '<FieldRef field="{}" ref="{}">'.format(
             self.field_name, self.doc_ref)

    @classmethod
    def from_string(cls, string):
        if cls.JOINER not in string:
            raise BaseLunrException('Malformed field ref string')
        field_ref, doc_ref = string.split(cls.JOINER, 1)
        return cls(doc_ref, field_ref, string)

    def __str__(self):
        if self._string_value is None:
            self._string_value = (
                self.field_name + self.JOINER + str(self.doc_ref))

        return self._string_value



from __future__ import unicode_literals

from copy import deepcopy


class MatchData:
    """Contains and collects metadata about a matching document.

    A single instance of lunr.MatchData is returned as part of every
    lunr.Index.Result.
    """

    def __init__(self, term=None, field=None, metadata=None):
        self.metadata = {}
        if term is not None:
            self.metadata[term] = {}
            if field is not None:
                self.metadata[term][field] = (
                    deepcopy(metadata) if metadata is not None else {})

    def __repr__(self):
        return '<MatchData "{}">'.format(
            ','.join(sorted(self.metadata.keys())))

    def combine(self, other):
        """An instance of lunr.MatchData will be created for every term that
        matches a document.

        However only one instance is required in a lunr.Index~Result. This
        method combines metadata from another instance of MatchData with this
        object's metadata.
        """
        for term in other.metadata.keys():
            if term not in self.metadata:
                self.metadata[term] = {}

            fields = other.metadata[term].keys()
            for field in fields:
                if field not in self.metadata[term]:
                    self.metadata[term][field] = {}

                keys = other.metadata[term][field].keys()
                for key in keys:
                    if key not in self.metadata[term][field]:
                        self.metadata[term][field][key] = (
                            other.metadata[term][field][key])
                    else:
                        self.metadata[term][field][key].extend(
                            other.metadata[term][field][key])

    def add(self, term, field, metadata):
        """Add metadata for a term/field pair to this instance of match data"""
        if term not in self.metadata:
            self.metadata[term] = {field: metadata}
            return

        if field not in self.metadata[term]:
            self.metadata[term][field] = metadata
            return

        for key in metadata.keys():
            if key in self.metadata[term][field]:
                self.metadata[term][field][key].extend(metadata[key])
            else:
                self.metadata[term][field][key] = metadata[key]

    def __eq__(self, other):
        return self.metadata == other.metadata



import os
import re
from setuptools import setup, find_packages


PATH = os.path.abspath(os.path.dirname(__file__))


def read_file(filepath):
    with open(filepath, 'r') as fd:
        return fd.read()


def find_version():
    version_path = os.path.join(PATH, 'lunr', '__init__.py')
    contents = read_file(version_path)
    version_string = contents[contents.index('__VERSION__'):]
    try:
        return re.match(
            r'.*__VERSION__ = [\'"]([\d\w\.]+)[\'"]', version_string).group(1)
    except AttributeError:
        raise RuntimeError("Unable to find version string.")


setup(
    name='lunr',
    version=find_version(),
    url='https://github.com/yeraydiazdiaz/lunr.py',
    license='MIT',
    description='A Python implementation of Lunr.js',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    author='Yeray Diaz Diaz',
    author_email='yeraydiazdiaz@gmail.com',
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    install_requires=[
        'future>=0.16.0',
        'six>=1.11.0',
    ],
    extras_require={
        'languages': ['nltk>=3.2.5']
    },
    keywords='lunr full text search',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Text Processing',
    ],
)



import pytest

from lunr import lunr
from lunr.query import Query, QueryPresence
from lunr.exceptions import QueryParseError


class TestSingleTermSearch:

    def test_one_match(self, index):
        results = index.search('scarlett')
        assert len(results) == 1
        assert results[0]['ref'] == 'c'
        assert set(
            results[0]['match_data'].metadata.keys()) == {'scarlett'}

    def test_no_match(self, index):
        results = index.search('foo')
        assert len(results) == 0

    def test_multiple_matches_sorts_by_relevance(self, index):
        results = index.search('plant')
        assert len(results) == 2
        assert results[0]['ref'] == 'b'
        assert results[1]['ref'] == 'c'

    def test_pipeline_processing_enabled(self, index):
        query = Query(index.fields)
        query.clause(
            term='study',  # stemmed to studi
            use_pipeline=True)
        results = index.query(query)

        assert len(results) == 2
        assert results[0]['ref'] == 'b'
        assert results[1]['ref'] == 'a'

    def test_pipeline_processing_disabled(self, index):
        query = Query(index.fields)
        query.clause(
            term='study',  # stemmed to studi
            use_pipeline=False)
        results = index.query(query)

        assert len(results) == 0

    def test_multiple_terms_all_terms_match(self, index):
        results = index.search('fellow candlestick')

        assert len(results) == 1
        assert results[0]['ref'] == 'a'
        metadata = results[0]['match_data'].metadata
        assert set(metadata.keys()) == {'fellow', 'candlestick'}
        assert set(metadata['fellow'].keys()) == {'body'}
        assert set(metadata['candlestick'].keys()) == {'body'}

    def test_one_term_matches(self, index):
        results = index.search('week foo')

        assert len(results) == 1
        assert results[0]['ref'] == 'c'
        metadata = results[0]['match_data'].metadata
        assert set(metadata.keys()) == {'week'}

    def test_duplicate_query_terms(self, index):
        index.search('fellow candlestick foo bar green plant fellow')

    def test_documents_with_all_terms_score_higher(self, index):
        results = index.search('candlestick green')

        assert len(results) == 3

        assert {r['ref'] for r in results} == {'a', 'b', 'c'}
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'candlestick', 'green'}
        assert set(results[1]['match_data'].metadata.keys()) == {'green'}
        assert set(results[2]['match_data'].metadata.keys()) == {'green'}

    def test_no_terms_match(self, index):
        results = index.search('foo bar')

        assert len(results) == 0

    def test_corpus_terms_are_stemmed(self, index):
        results = index.search('water')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}

    def test_field_scoped_terms(self, index):
        results = index.search('title:plant')

        assert len(results) == 1
        assert results[0]['ref'] == 'b'
        assert set(results[0]['match_data'].metadata.keys()) == {'plant'}

    def test_field_scoped_no_matching_terms(self, index):
        results = index.search('title:candlestick')

        assert len(results) == 0


class TestSearchWildcardTrailing:

    def test_matching_no_matches(self, index):
        results = index.search('fo*')

        assert len(results) == 0

    def test_matching_one_match(self, index):
        results = index.search('candle*')

        assert len(results) == 1
        assert results[0]['ref'] == 'a'
        assert set(
            results[0]['match_data'].metadata.keys()) == {'candlestick'}

    def test_matching_multiple_terms_match(self, index):
        results = index.search('pl*')

        assert len(results) == 2

        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'plumb', 'plant'}
        assert set(
            results[1]['match_data'].metadata.keys()
        ) == {'plumb', 'plant'}


class TestSearchWildcardLeading:

    def test_matching_no_matches(self, index):
        results = index.search('*oo')

        assert len(results) == 0

    def test_matching_multiple_terms_match(self, index):
        results = index.search('*ant')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(results[0]['match_data'].metadata.keys()) == {'plant'}
        assert set(results[1]['match_data'].metadata.keys()) == {'plant'}


class TestSearchWildcardContained:

    def test_matching_no_matches(self, index):
        results = index.search('f*o')
        assert len(results) == 0

    def test_matching_multiple_terms_match(self, index):
        results = index.search('pl*nt')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(results[0]['match_data'].metadata.keys()) == {'plant'}
        assert set(results[1]['match_data'].metadata.keys()) == {'plant'}


class TestEditDistance:

    def test_edit_distance_no_results(self, index):
        results = index.search('foo~1')
        assert len(results) == 0

    def test_edit_distance_two_results(self, index):
        results = index.search('plont~1')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(results[0]['match_data'].metadata.keys()) == {'plant'}
        assert set(results[1]['match_data'].metadata.keys()) == {'plant'}


class TestSearchByField:

    def test_search_by_field_unknown_field(self, index):
        with pytest.raises(QueryParseError):
            index.search('unknown-field:plant')

    def test_search_by_field_no_results(self, index):
        results = index.search('title:candlestick')
        assert len(results) == 0

    def test_search_by_field_results(self, index):
        results = index.search('title:plant')

        assert len(results) == 1
        assert results[0]['ref'] == 'b'
        assert set(results[0]['match_data'].metadata.keys()) == {'plant'}


class TestTermBoosts:

    def test_term_boosts_no_results(self, index):
        results = index.search('foo^10')
        assert len(results) == 0

    def test_term_boosts_results(self, index):
        results = index.search('scarlett candlestick^5')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'a', 'c'}
        assert set(
            results[0]['match_data'].metadata.keys()) == {'candlestick'}
        assert set(results[1]['match_data'].metadata.keys()) == {'scarlett'}


class TestTypeaheadStyleSearch:

    def test_typeahead_no_results(self, index):
        query = Query(index.fields)
        query.clause('xyz', boost=100, use_pipeline=True)
        query.clause(
            'xyz', boost=10, use_pipeline=False,
            wildcard=Query.WILDCARD_TRAILING)
        query.clause('xyz', boost=1, edit_distance=1)

        results = index.query(query)
        assert len(results) == 0

    def test_typeahead_results(self, index):
        query = Query(index.fields)
        query.clause('pl', boost=100, use_pipeline=True)
        query.clause(
            'pl', boost=10, use_pipeline=False,
            wildcard=Query.WILDCARD_TRAILING)
        query.clause('pl', boost=1, edit_distance=1)

        results = index.query(query)

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(
            results[0]['match_data'].metadata.keys()) == {'plumb', 'plant'}
        assert set(
            results[1]['match_data'].metadata.keys()) == {'plumb', 'plant'}


class TestTermPresence:

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_prohibited_match_excludes_prohibited_result(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('candlestick', presence=QueryPresence.PROHIBITED)
            query.term('green', presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search('-candlestick green')

        assert len(results) == 2
        assert {r['ref'] for r in results} == {'b', 'c'}
        assert set(
            results[0]['match_data'].metadata.keys()) == {'green'}
        assert set(
            results[1]['match_data'].metadata.keys()) == {'green'}

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_only_prohibited_match_yields_no_results(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('green', presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search('-green')

        assert len(results) == 0

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_negated_query_no_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('qwertyuiop', presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search('-qwertyuiop')

        assert len(results) == 3

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_negated_query_some_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('plant', presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search('-plant')

        assert len(results) == 1
        assert results[0]['score'] == 0

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_field_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term(
                'plant', presence=QueryPresence.PROHIBITED, fields=['title'])
            query.term('plumb', presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search('-title:plant plumb')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()) == {'plumb'}

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_required_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('candlestick', presence=QueryPresence.REQUIRED)
            query.term('green', presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search('+candlestick green')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'candlestick', 'green'}

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_two_required_matches_yields_no_results(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('mustard', presence=QueryPresence.REQUIRED)
            query.term('plant', presence=QueryPresence.REQUIRED)
            results = index.query(query)
        else:
            results = index.search('+mustard +plant')

        assert len(results) == 0

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_required_term_not_matching_yields_no_results(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('qwertyuiop', presence=QueryPresence.REQUIRED)
            query.term('green', presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search('+qwertyuiop green')

        assert len(results) == 0

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_required_term_on_field_matches(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term(
                'plant', presence=QueryPresence.REQUIRED, fields=['title'])
            query.term('green', presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search('+title:plant green')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'plant', 'green'}
        assert results[0]['ref'] == 'b'

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_required_terms_on_field_and_non_field_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term(
                'plant', presence=QueryPresence.REQUIRED, fields=['title'])
            query.term('green', presence=QueryPresence.REQUIRED)
            results = index.query(query)
        else:
            results = index.search('+title:plant +green')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'plant', 'green'}
        assert results[0]['ref'] == 'b'

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_required_terms_on_different_fields_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term(
                'plant', presence=QueryPresence.REQUIRED, fields=['title'])
            query.term(
                'study', presence=QueryPresence.REQUIRED, fields=['body'])
            results = index.query(query)
        else:
            results = index.search('+title:plant +body:study')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'studi', 'plant'}
        assert results[0]['ref'] == 'b'

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_combined_required_optional_and_prohibited_match(
            self, index, query_or_search):
        if query_or_search == 'query':
            query = Query(index.fields)
            query.term('plant', presence=QueryPresence.REQUIRED)
            query.term('green', presence=QueryPresence.OPTIONAL)
            query.term('office', presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search('+plant green -office')

        assert len(results) == 1
        assert set(
            results[0]['match_data'].metadata.keys()
        ) == {'green', 'plant'}
        assert results[0]['ref'] == 'b'


class TestBuildTimeFieldBoost:

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_no_query_boosts_build_boost_ranks_higher(
            self, documents, query_or_search):
        idx = lunr(
            ref='id',
            fields=['title', {
                'field_name': 'body',
                'boost': 10,
            }],
            documents=documents
        )

        if query_or_search == 'search':
            results = idx.search('professor')
        else:
            query = Query(idx.fields)
            query.term('professor')
            results = idx.query(query)

        assert results[0]['ref'] == 'b'


class TestBuildTimeDocumentBoost:

    @pytest.mark.parametrize('query_or_search', ['query', 'search'])
    def test_no_query_boosts_document_boost_ranks_higher(
                self, documents, query_or_search):
        documents_with_boost = [
            (document, {'boost': 10 if document['id'] == 'c' else 1})
            for document in documents
        ]
        idx = lunr(
            ref='id',
            fields=('title', 'body'),
            documents=documents_with_boost
        )

        if query_or_search == 'search':
            results = idx.search('plumb')
        else:
            query = Query(idx.fields)
            query.term('plumb')
            results = idx.query(query)

        assert results[0]['ref'] == 'c'



from __future__ import unicode_literals

from builtins import str

from lunr.pipeline import Pipeline

WORDS = [
    'a',
    'able',
    'about',
    'across',
    'after',
    'all',
    'almost',
    'also',
    'am',
    'among',
    'an',
    'and',
    'any',
    'are',
    'as',
    'at',
    'be',
    'because',
    'been',
    'but',
    'by',
    'can',
    'cannot',
    'could',
    'dear',
    'did',
    'do',
    'does',
    'either',
    'else',
    'ever',
    'every',
    'for',
    'from',
    'get',
    'got',
    'had',
    'has',
    'have',
    'he',
    'her',
    'hers',
    'him',
    'his',
    'how',
    'however',
    'i',
    'if',
    'in',
    'into',
    'is',
    'it',
    'its',
    'just',
    'least',
    'let',
    'like',
    'likely',
    'may',
    'me',
    'might',
    'most',
    'must',
    'my',
    'neither',
    'no',
    'nor',
    'not',
    'of',
    'off',
    'often',
    'on',
    'only',
    'or',
    'other',
    'our',
    'own',
    'rather',
    'said',
    'say',
    'says',
    'she',
    'should',
    'since',
    'so',
    'some',
    'than',
    'that',
    'the',
    'their',
    'them',
    'then',
    'there',
    'these',
    'they',
    'this',
    'tis',
    'to',
    'too',
    'twas',
    'us',
    'wants',
    'was',
    'we',
    'were',
    'what',
    'when',
    'where',
    'which',
    'while',
    'who',
    'whom',
    'why',
    'will',
    'with',
    'would',
    'yet',
    'you',
    'your'
]


def generate_stop_word_filter(stop_words, language=None):
    """Builds a stopWordFilter function from the provided list of stop words.

    The built in `stop_word_filter` is built using this factory and can be used
    to generate custom `stop_word_filter` for applications or non English
    languages.
    """
    # TODO: do we need this dict?
    words = {word: word for word in stop_words}

    def stop_word_filter(token, i=None, tokens=None):
        if token and words.get(str(token)) != str(token):
            return token

    # camelCased for for compatibility with lunr.js
    label = (
        'stopWordFilter-{}'.format(language) if language is not None
        else 'stopWordFilter')
    Pipeline.register_function(stop_word_filter, label)
    return stop_word_filter


stop_word_filter = generate_stop_word_filter(WORDS)



from lunr.query_lexer import QueryLexer


def _lex(string):
    lexer = QueryLexer(string)
    lexer.run()
    return lexer


class TestQueryLexer:

    def test_single_term_produces_one_lexeme(self):
        lexer = _lex('foo')
        assert len(lexer.lexemes) == 1
        lexeme = lexer.lexemes[0]
        assert lexeme['type'] == QueryLexer.TERM
        assert lexeme['string'] == 'foo'
        assert lexeme['start'] == 0
        assert lexeme['end'] == 3

    def test_term_escape_character(self):
        lexer = _lex('foo\\:bar')
        assert len(lexer.lexemes) == 1
        lexeme = lexer.lexemes[0]
        assert lexeme['type'] == QueryLexer.TERM
        assert lexeme['string'] == 'foo:bar'
        assert lexeme['start'] == 0
        assert lexeme['end'] == 8

    def test_multiple_terms(self):
        lexer = _lex('foo bar')
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes
        assert foo_lexeme['type'] == bar_lexeme['type'] == QueryLexer.TERM
        assert foo_lexeme['string'] == 'foo'
        assert bar_lexeme['string'] == 'bar'
        assert foo_lexeme['start'] == 0
        assert bar_lexeme['start'] == 4
        assert foo_lexeme['end'] == 3
        assert bar_lexeme['end'] == 7

    def test_separator_length_greater_than_one(self):
        lexer = _lex('foo    bar')
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes
        assert foo_lexeme['type'] == bar_lexeme['type'] == QueryLexer.TERM
        assert foo_lexeme['string'] == 'foo'
        assert bar_lexeme['string'] == 'bar'
        assert foo_lexeme['start'] == 0
        assert bar_lexeme['start'] == 7
        assert foo_lexeme['end'] == 3
        assert bar_lexeme['end'] == 10

    def test_hyphen_is_considered_a_separator(self):
        lexer = _lex('foo-bar')
        assert len(lexer.lexemes) == 2

    def test_term_with_field(self):
        lexer = _lex('title:foo')
        assert len(lexer.lexemes) == 2
        field_lexeme, term_lexeme = lexer.lexemes
        assert field_lexeme['type'] == QueryLexer.FIELD
        assert term_lexeme['type'] == QueryLexer.TERM
        assert field_lexeme['string'] == 'title'
        assert term_lexeme['string'] == 'foo'
        assert field_lexeme['start'] == 0
        assert term_lexeme['start'] == 6
        assert field_lexeme['end'] == 5
        assert term_lexeme['end'] == 9

    def test_term_with_field_with_escape_character(self):
        lexer = _lex('ti\\:tle:foo')
        assert len(lexer.lexemes) == 2
        field_lexeme, term_lexeme = lexer.lexemes
        assert field_lexeme['type'] == QueryLexer.FIELD
        assert term_lexeme['type'] == QueryLexer.TERM
        assert field_lexeme['string'] == 'ti:tle'
        assert term_lexeme['string'] == 'foo'
        assert field_lexeme['start'] == 0
        assert term_lexeme['start'] == 8
        assert field_lexeme['end'] == 7
        assert term_lexeme['end'] == 11

    def test_term_with_edit_distance(self):
        lexer = _lex('foo~2')
        assert len(lexer.lexemes) == 2
        term_lexeme, edit_distance_lexeme = lexer.lexemes
        assert term_lexeme['type'] == QueryLexer.TERM
        assert edit_distance_lexeme['type'] == QueryLexer.EDIT_DISTANCE
        assert term_lexeme['string'] == 'foo'
        assert edit_distance_lexeme['string'] == '2'
        assert term_lexeme['start'] == 0
        assert edit_distance_lexeme['start'] == 4
        assert term_lexeme['end'] == 3
        assert edit_distance_lexeme['end'] == 5

    def test_term_with_boost(self):
        lexer = _lex('foo^10')
        assert len(lexer.lexemes) == 2
        term_lexeme, boost_lexeme = lexer.lexemes
        assert term_lexeme['type'] == QueryLexer.TERM
        assert boost_lexeme['type'] == QueryLexer.BOOST
        assert term_lexeme['string'] == 'foo'
        assert boost_lexeme['string'] == '10'
        assert term_lexeme['start'] == 0
        assert boost_lexeme['start'] == 4
        assert term_lexeme['end'] == 3
        assert boost_lexeme['end'] == 6

    def test_term_with_field_boost_and_edit_distance(self):
        lexer = _lex('title:foo^10~5')
        assert len(lexer.lexemes) == 4
        field_lexeme, term_lexeme, boost_lexeme, edit_distance_lexeme = (
            lexer.lexemes)
        assert field_lexeme['type'] == QueryLexer.FIELD
        assert term_lexeme['type'] == QueryLexer.TERM
        assert boost_lexeme['type'] == QueryLexer.BOOST
        assert edit_distance_lexeme['type'] == QueryLexer.EDIT_DISTANCE

        assert field_lexeme['string'] == 'title'
        assert term_lexeme['string'] == 'foo'
        assert boost_lexeme['string'] == '10'
        assert edit_distance_lexeme['string'] == '5'

        assert field_lexeme['start'] == 0
        assert term_lexeme['start'] == 6
        assert boost_lexeme['start'] == 10
        assert edit_distance_lexeme['start'] == 13

        assert field_lexeme['end'] == 5
        assert term_lexeme['end'] == 9
        assert boost_lexeme['end'] == 12
        assert edit_distance_lexeme['end'] == 14

    def test_single_term_with_hyphen_produces_two_lexemes(self):
        """Embedded hyphens should not be confused with presence operators."""
        lexer = _lex('foo-bar')
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes

        assert foo_lexeme['type'] == QueryLexer.TERM
        assert foo_lexeme['string'] == 'foo'
        assert foo_lexeme['start'] == 0
        assert foo_lexeme['end'] == 3

        assert bar_lexeme['type'] == QueryLexer.TERM
        assert bar_lexeme['string'] == 'bar'
        assert bar_lexeme['start'] == 4
        assert bar_lexeme['end'] == 7

    def test_single_term_with_presence_produces_two_lexemes(self):
        lexer = _lex('+foo')
        assert len(lexer.lexemes) == 2
        presence_lexeme, term_lexeme = lexer.lexemes

        assert presence_lexeme['type'] == QueryLexer.PRESENCE
        assert presence_lexeme['string'] == '+'
        assert presence_lexeme['start'] == 0
        assert presence_lexeme['end'] == 1

        assert term_lexeme['type'] == QueryLexer.TERM
        assert term_lexeme['string'] == 'foo'
        assert term_lexeme['start'] == 1
        assert term_lexeme['end'] == 4

    def test_multiple_terms_with_presence_produces_four_lexemes(self):
        lexer = _lex('+foo +bar')
        assert len(lexer.lexemes) == 4
        (foo_presence_lexeme, foo_term_lexeme,
         bar_presence_lexeme, bar_term_lexeme) = lexer.lexemes

        assert foo_term_lexeme['type'] == QueryLexer.TERM
        assert foo_term_lexeme['string'] == 'foo'
        assert foo_term_lexeme['start'] == 1
        assert foo_term_lexeme['end'] == 4

        assert foo_presence_lexeme['type'] == QueryLexer.PRESENCE
        assert foo_presence_lexeme['string'] == '+'
        assert foo_presence_lexeme['start'] == 0
        assert foo_presence_lexeme['end'] == 1

        assert bar_term_lexeme['type'] == QueryLexer.TERM
        assert bar_term_lexeme['string'] == 'bar'
        assert bar_term_lexeme['start'] == 6
        assert bar_term_lexeme['end'] == 9

        assert bar_presence_lexeme['type'] == QueryLexer.PRESENCE
        assert bar_presence_lexeme['string'] == '+'
        assert bar_presence_lexeme['start'] == 5
        assert bar_presence_lexeme['end'] == 6

    def test_multiple_terms_with_presence_and_fuzz(self):
        lexer = _lex('+foo~1 +bar')
        assert len(lexer.lexemes) == 5

        (foo_presence_lexeme, foo_term_lexeme, foo_fuzz_lexeme,
         bar_presence_lexeme, bar_term_lexeme) = lexer.lexemes

        assert foo_presence_lexeme['type'] == QueryLexer.PRESENCE
        assert foo_presence_lexeme['string'] == '+'
        assert foo_presence_lexeme['start'] == 0
        assert foo_presence_lexeme['end'] == 1

        assert foo_term_lexeme['type'] == QueryLexer.TERM
        assert foo_term_lexeme['string'] == 'foo'
        assert foo_term_lexeme['start'] == 1
        assert foo_term_lexeme['end'] == 4

        assert foo_fuzz_lexeme['type'] == QueryLexer.EDIT_DISTANCE
        assert foo_fuzz_lexeme['string'] == '1'
        assert foo_fuzz_lexeme['start'] == 5
        assert foo_fuzz_lexeme['end'] == 6

        assert bar_presence_lexeme['type'] == QueryLexer.PRESENCE
        assert bar_presence_lexeme['string'] == '+'
        assert bar_presence_lexeme['start'] == 7
        assert bar_presence_lexeme['end'] == 8

        assert bar_term_lexeme['type'] == QueryLexer.TERM
        assert bar_term_lexeme['string'] == 'bar'
        assert bar_term_lexeme['start'] == 8
        assert bar_term_lexeme['end'] == 11



import pytest

from lunr.query import Query, QueryPresence
from lunr.query_parser import QueryParser
from lunr.exceptions import QueryParseError


def parse(q):
    query = Query(['title', 'body'])
    parser = QueryParser(q, query)

    parser.parse()
    return query.clauses


class TestQueryParser:

    def test_parse_empty_string(self):
        clauses = parse('')
        assert len(clauses) == 0

    def test_parse_single_term(self):
        clauses = parse('foo')
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == 'foo'
        assert clause.fields == ['title', 'body']
        assert clause.use_pipeline is True
        assert clause.presence is QueryPresence.OPTIONAL

    def test_parse_single_term_uppercase(self):
        clauses = parse('FOO')
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == 'foo'
        assert clause.fields == ['title', 'body']
        assert clause.use_pipeline is True

    def test_parse_single_term_with_wildcard(self):
        clauses = parse('fo*')
        assert len(clauses) == 1
        clause = clauses[0]
        assert clause.term == 'fo*'
        assert clause.use_pipeline is False

    def test_multiple_terms(self):
        clauses = parse('foo bar')
        assert len(clauses) == 2
        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

    def test_term_with_presence_required_adds_required_clause(self):
        clauses = parse('+foo')
        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].boost == 1
        assert clauses[0].fields == ['title', 'body']
        assert clauses[0].presence == QueryPresence.REQUIRED

    def test_term_with_presence_required_adds_prohibited_clause(self):
        clauses = parse('-foo')
        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].boost == 1
        assert clauses[0].fields == ['title', 'body']
        assert clauses[0].presence == QueryPresence.PROHIBITED

    def test_term_scoped_by_field_with_presence_required(self):
        clauses = parse('+title:foo')
        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].boost == 1
        assert clauses[0].fields == ['title']
        assert clauses[0].presence == QueryPresence.REQUIRED

    def test_term_scoped_by_field_with_presence_prohibited(self):
        clauses = parse('-title:foo')
        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].boost == 1
        assert clauses[0].fields == ['title']
        assert clauses[0].presence == QueryPresence.PROHIBITED

    def test_multiple_terms_with_presence_creates_two_clauses(self):
        clauses = parse('+foo +bar')
        assert len(clauses) == 2
        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'
        assert clauses[0].presence == QueryPresence.REQUIRED
        assert clauses[1].presence == QueryPresence.REQUIRED

    def test_unknown_field(self):
        with pytest.raises(QueryParseError):
            parse('unknown:foo')

    def test_field_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse('title:')

    def test_field_twice(self):
        with pytest.raises(QueryParseError):
            parse('title:title:')

    def test_term_with_field(self):
        clauses = parse('title:foo')
        assert len(clauses) == 1
        assert clauses[0].fields == ['title']

    def test_uppercase_field_with_uppercase_term(self):
        query = Query(['TITLE'])
        parser = QueryParser('TITLE:FOO', query)

        parser.parse()
        clauses = query.clauses

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['TITLE']

    def test_multiple_terms_scoped_to_different_fields(self):
        clauses = parse('title:foo body:bar')

        assert len(clauses) == 2
        assert clauses[0].fields == ['title']
        assert clauses[1].fields == ['body']

        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

    def test_single_term_with_edit_distance(self):
        clauses = parse('foo~2')

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['title', 'body']
        assert clauses[0].edit_distance == 2

    def test_multiple_terms_with_edit_distance(self):
        clauses = parse('foo~2 bar~3')

        assert len(clauses) == 2
        assert clauses[0].fields == ['title', 'body']
        assert clauses[1].fields == ['title', 'body']

        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

        assert clauses[0].edit_distance == 2
        assert clauses[1].edit_distance == 3

    def test_single_term_scoped_to_field_with_edit_distance(self):
        clauses = parse('title:foo~2')

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['title']
        assert clauses[0].edit_distance == 2

    def test_non_numeric_edit_distance(self):
        with pytest.raises(QueryParseError):
            parse('foo~a')

    def test_edit_distance_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse('~2')

    def test_single_term_with_boost(self):
        clauses = parse('foo^2')

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['title', 'body']
        assert clauses[0].boost == 2

    def test_non_numeric_boost(self):
        with pytest.raises(QueryParseError):
            parse('foo^a')

    def test_boost_without_a_term(self):
        with pytest.raises(QueryParseError):
            parse('^2')

    def test_multiple_terms_with_boost(self):
        clauses = parse('foo^2 bar^3')

        assert len(clauses) == 2
        assert clauses[0].fields == ['title', 'body']
        assert clauses[1].fields == ['title', 'body']

        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

        assert clauses[0].boost == 2
        assert clauses[1].boost == 3

    def test_term_scoped_by_field_with_boost(self):
        clauses = parse('title:foo^2')

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['title']
        assert clauses[0].boost == 2

    def test_term_with_boost_and_edit_distance(self):
        clauses = parse('foo^2~3')

        assert len(clauses) == 1
        assert clauses[0].term == 'foo'
        assert clauses[0].fields == ['title', 'body']
        assert clauses[0].edit_distance == 3
        assert clauses[0].boost == 2

    def test_edit_distance_followed_by_presence(self):
        clauses = parse('foo~10 +bar')

        assert len(clauses) == 2

        assert clauses[0].fields == ['title', 'body']
        assert clauses[1].fields == ['title', 'body']

        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

        assert clauses[0].edit_distance == 10
        assert clauses[1].edit_distance == 0

        assert clauses[0].presence == QueryPresence.OPTIONAL
        assert clauses[1].presence == QueryPresence.REQUIRED

    def test_boost_followed_by_presence(self):
        clauses = parse('foo^10 +bar')

        assert len(clauses) == 2

        assert clauses[0].fields == ['title', 'body']
        assert clauses[1].fields == ['title', 'body']

        assert clauses[0].term == 'foo'
        assert clauses[1].term == 'bar'

        assert clauses[0].boost == 10
        assert clauses[1].boost == 1

        assert clauses[0].presence == QueryPresence.OPTIONAL
        assert clauses[1].presence == QueryPresence.REQUIRED
